"""yt-dlg __main__ file.

__main__ file is a python 'executable' file which calls the youtube_dl_gui.app
main() function in order to start the app. It can be used to start
the app from the package directory OR it can be used to start the app
from a different directory after you have installed the youtube_dl_gui
package.

Example:
    In order to run the app from the package directory.

        $ cd <package directory>
        $ python __main__.py

    In order to run the app AFTER you have installed the package using
    setup.py.

        $ yt-dlg

"""


import sys
from pathlib import Path

if __package__ is None and not hasattr(sys, "frozen"):
    # direct call of __main__.py
    PATH = Path(__file__).resolve().parent
    sys.path.insert(0, str(PATH))

from youtube_dl_gui.app import main  # type: ignore[attr-defined]

sys.exit(main())



"""
Author: Sotiris Papadopoulos <ytubedlg@gmail.com>
Last-Revision: 2017-04-19

Script to automatically check PO files

"""


import argparse
import logging
import os
import sys
from datetime import datetime, timedelta, tzinfo
from time import sleep

try:
    import google_translate
    import polib
except ImportError as error:
    print(error)
    sys.exit(1)


WTIME = 2.0  # Time in seconds to wait between requests to avoid ban

PACKAGE = "youtube_dl_gui"

PO_FILENAME = f"{PACKAGE}.po"

LOCALE_PATH_TMPL = os.path.join(PACKAGE, "locale", "{lang}", "LC_MESSAGES", PO_FILENAME)


logging.basicConfig(level=logging.ERROR)


def parse():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Script to automatically check PO files"
    )

    parser.add_argument("language", help="language of the PO file to check")

    parser.add_argument(
        "-w",
        "--werror",
        action="store_true",
        help="treat all warning messages as errors",
    )
    parser.add_argument(
        "-o",
        "--only-headers",
        action="store_true",
        help="check only the PO file headers",
    )
    parser.add_argument(
        "-n",
        "--no-translate",
        action="store_true",
        help="do not use the translator to check 'msgstr' fields",
    )
    parser.add_argument(
        "-t",
        "--tlang",
        help="force a different language on the translator than the one given",
    )

    return parser.parse_args()


class UTC_Offset_Timezone(tzinfo):

    """Class that represents a UTC offset in the format +/-0000."""

    def __init__(self, offset_string):
        self.offset = timedelta(seconds=UTC_Offset_Timezone.parse_offset(offset_string))

    def utcoffset(self, dt):
        return self.offset + self.dst(dt)

    def dst(self, dt):
        return timedelta(0)

    @staticmethod
    def parse_offset(offset_string):
        """Parse the offset string into seconds."""

        if len(offset_string) != 5:
            raise ValueError(f"Invalid length for offset string ({offset_string})")

        hours = offset_string[1:3]
        minutes = offset_string[3:5]

        offset = int(hours) * 3600 + int(minutes) * 60

        if offset_string[0] == "-":
            return -1 * offset

        return offset


def parse_date(date_string):
    """Parse date string into an aware datetime object."""

    # Just a small list with the most common timezones
    offset_list = [
        ("JST", "0900"),
        ("EEST", "0300"),
        ("EET", "0200"),
        ("GMT", "0000"),
        ("UTC", "0000"),
    ]

    # Replace all the timezones with the offset
    for item in offset_list:
        timezone, offset = item

        date_string = date_string.replace(timezone, offset)

    datetime_string = date_string[:16]
    offset_string = date_string[16:]

    naive_date = datetime.strptime(datetime_string, "%Y-%m-%d %H:%M")

    # Create & return an aware datetime object based on the offset
    return naive_date.replace(tzinfo=UTC_Offset_Timezone(offset_string))


# Print helpers


def my_print(msg, char="*", value=None, exit=False):
    """Print 'msg', debug 'value' and exit if 'exit' is True."""
    print(f"[{char}] {msg}")

    if value is not None:
        print(f'\tvalue= "{value}"')

    if exit:
        sys.exit(1)


def perror(msg, value=None):
    my_print(msg, "-", value, True)


def pwarn(msg, value=None, exit=False):
    my_print(msg, "!", value, exit)


def pinfo(msg):
    my_print(msg)


#############################


def main(args):
    os.chdir("..")

    # setup
    pot_file_path = LOCALE_PATH_TMPL.format(lang="en_US")
    po_file_path = LOCALE_PATH_TMPL.format(lang=args.language)

    if not os.path.exists(pot_file_path):
        perror("Failed to locate POT file, exiting...", pot_file_path)

    if not os.path.exists(po_file_path):
        perror("Failed to locate PO file, exiting...", po_file_path)

    pot_file = polib.pofile(pot_file_path)
    po_file = polib.pofile(po_file_path)

    # check headers
    pinfo("Checking PO headers")

    pot_headers = pot_file.metadata
    po_headers = po_file.metadata

    if pot_headers["Project-Id-Version"] != po_headers["Project-Id-Version"]:
        pwarn("'Project-Id-Version' headers do not match", exit=args.werror)

    if pot_headers["POT-Creation-Date"] != po_headers["POT-Creation-Date"]:
        pwarn("'POT-Creation-Date' headers do not match", exit=args.werror)

    po_creation_date = parse_date(po_headers["POT-Creation-Date"])
    po_revision_date = parse_date(po_headers["PO-Revision-Date"])

    # Aware datetimes convert to UTC automatically when comparing
    if po_revision_date <= po_creation_date:
        pwarn("PO file seems outdated", exit=args.werror)

    if "Language" in po_headers and po_headers["Language"] != args.language:
        pwarn(
            "'Language' header does not match with the given language",
            po_headers["Language"],
            args.werror,
        )

    pinfo("Last-Translator: {}".format(po_headers["Last-Translator"]))

    # check translations
    if args.only_headers:
        sys.exit(0)

    pinfo("Checking translations, this might take a while...")

    pot_msgid = [entry.msgid for entry in pot_file]
    po_msgid = [entry.msgid for entry in po_file]

    not_translated = []
    same_msgstr = []
    with_typo = []
    verify_trans = []
    fuzzy_trans = po_file.fuzzy_entries()

    # lists to hold reports
    missing_msgid = [msgid for msgid in pot_msgid if msgid not in po_msgid]
    # Init translator only if the '--no-translate' flag is NOT set
    translator = None
    if not args.no_translate:
        translator = google_translate.GoogleTranslator(
            timeout=5.0, retries=2, wait_time=WTIME
        )

        # Set source language for GoogleTranslator
        if args.tlang is not None:
            src_lang = args.tlang
            pinfo(f"Forcing '{src_lang}' as the translator's source language")
        else:
            # Get a valid source language for Google
            # for example convert 'ar_SA' to 'ar' or 'zh_CN' to 'zh-CN'
            src_lang = args.language

            if src_lang not in translator._lang_dict:
                src_lang = src_lang.replace("_", "-")

            if src_lang not in translator._lang_dict:
                src_lang = src_lang.split("-")[0]

    # Keep entries that need further analysis using the translator
    further_analysis = []

    for entry in po_file:
        if not entry.translated():
            not_translated.append(entry)

        elif entry.msgid == entry.msgstr:
            same_msgstr.append(entry)

        else:
            further_analysis.append(entry)

    if translator is not None and further_analysis:
        # eta = (items_to_analyze * (WTIME + avg_ms)) - WTIME
        # We subtract WTIME at the end because there is no wait for the last item on the list
        # avg_msg = 200ms
        eta_seconds = (len(further_analysis) * (WTIME + 0.2)) - WTIME
        eta_seconds = int(round(eta_seconds))

        eta = timedelta(seconds=eta_seconds)
        pinfo(f"Approximate time to check translations online: {eta}")

        # Pass translations as a list since GoogleTranslator can handle them
        words_dict = translator.get_info_dict(
            [entry.msgstr for entry in further_analysis], "en", src_lang
        )

        for index, word_dict in enumerate(words_dict):
            # Get the corresponding POEntry since the words_dict does not contain those
            entry = further_analysis[index]

            if word_dict is not None:
                if word_dict["has_typo"]:
                    with_typo.append(entry)

                if word_dict["translation"].lower() != entry.msgid.lower():
                    found = any(
                        entry.msgid.lower() in word_dict["extra"][key].keys()
                        for key in word_dict["extra"]
                    )

                    if not found:
                        verify_trans.append((entry, word_dict["translation"]))

    # time to report
    print("=" * 25 + "Report" + "=" * 25)

    if missing_msgid:
        print("Missing msgids")

        for msgid in missing_msgid:
            print(f'  "{msgid}"')

    if not_translated:
        print("Not translated")

        for entry in not_translated:
            print(f'  line: {entry.linenum} msgid: "{entry.msgid}"')

    if same_msgstr:
        print("Same msgstr")

        for entry in same_msgstr:
            print(f'  line: {entry.linenum} msgid: "{entry.msgid}"')

    if with_typo:
        print("With typo")

        for entry in with_typo:
            print(
                '  line: {} msgid: "{}" msgstr: "{}"'.format(
                    entry.linenum, entry.msgid, entry.msgstr
                )
            )

    if verify_trans:
        print("Verify translation")

        for item in verify_trans:
            entry, translation = item
            print(
                '  line: {} msgid: "{}" trans: "{}"'.format(
                    entry.linenum, entry.msgid, translation
                )
            )

    if fuzzy_trans:
        print("Fuzzy translations")

        for entry in fuzzy_trans:
            print(f'  line: {entry.linenum} msgid: "{entry.msgid}"')

    total = (
        len(missing_msgid)
        + len(not_translated)
        + len(same_msgstr)
        + len(with_typo)
        + len(verify_trans)
        + len(fuzzy_trans)
    )

    print("")
    print(f"Missing msgids\t\t: {len(missing_msgid)}")
    print(f"Not translated\t\t: {len(not_translated)}")
    print(f"Same msgstr\t\t: {len(same_msgstr)}")
    print(f"With typo\t\t: {len(with_typo)}")
    print(f"Verify translation\t: {len(verify_trans)}")
    print(f"Fuzzy translations\t: {len(fuzzy_trans)}")
    print(f"Total\t\t\t: {total}")
    print("")
    print(f"Total entries\t\t: {len(po_file)}")


if __name__ == "__main__":
    try:
        main(parse())
    except KeyboardInterrupt:
        print("KeyboardInterrupt")



"""
Author: Sotiris Papadopoulos <ytubedlg@gmail.com>
Last-Revision: 2017-01-30

Script to add support for a new language

Usage   : ./new-locale.py <language_code>
Example : ./new-locale.py en_US

"""

import os
import shutil
import sys

PACKAGE = "youtube_dl_gui"

LOCALE_PATH_TMPL = os.path.join(PACKAGE, "locale", "{lang}", "LC_MESSAGES")

PO_FILE_TMPL = os.path.join("{parent_dir}", "youtube_dl_gui.po")


def error(msg):
    print(f"[-]{msg}")
    sys.exit(1)


def output(msg):
    print(f"[*]{msg}")


def manage_directory():
    """Allow script calls from the 'devscripts' dir and the package dir."""
    if os.path.basename(os.getcwd()) == "devscripts":
        os.chdir("..")


def main(lang_code):
    manage_directory()

    target_dir = LOCALE_PATH_TMPL.format(lang=lang_code)
    default_dir = LOCALE_PATH_TMPL.format(lang="en_US")

    target_po = PO_FILE_TMPL.format(parent_dir=target_dir)
    source_po = PO_FILE_TMPL.format(parent_dir=default_dir)

    if os.path.exists(target_dir):
        error(f"Locale '{lang_code}' already exists, exiting...")

    output(f"Creating directory: '{target_dir}'")
    os.makedirs(target_dir)

    output(f"Creating PO file: '{target_po}'")
    shutil.copy(source_po, target_po)

    output("Done")


if __name__ == "__main__":
    if len(sys.argv) == 2:
        main(sys.argv[1])
    else:
        print(f"Usage   : {sys.argv[0]} <language_code>")
        print(f"Example : {sys.argv[0]} en_US")



"""Python module to download videos.

This module contains the actual downloaders responsible
for downloading the video files.

"""

from __future__ import annotations

import contextlib
import os
import re
import signal
import subprocess
from pathlib import Path
from queue import Queue
from threading import Thread
from time import sleep
from typing import IO, Any, Callable

from .utils import IS_WINDOWS, get_encoding


# noinspection PyUnresolvedReferences
class PipeReader(Thread):
    """Helper class to avoid deadlocks when reading from subprocess pipes.

    This class uses python threads and queues in order to read from subprocess
    pipes in an asynchronous way.

    Attributes:
        WAIT_TIME (float): Time in seconds to sleep.

    Args:
        queue (Queue): Python queue to store the output of the subprocess.

    Warnings:
        All the operations are based on 'str' types.

    """

    WAIT_TIME = 0.1

    def __init__(self, queue: Queue):
        super().__init__()
        self._filedescriptor: IO[str] | None = None
        self._running: bool = True
        self._queue: Queue[str] = queue
        self.start()

    def run(self) -> None:
        # Flag to ignore specific lines
        ignore_line: bool = False

        while self._running:
            if self._filedescriptor is not None and not self._filedescriptor.closed:
                pipedata: str = self._filedescriptor.read()

                for line in pipedata.splitlines():
                    # Ignore ffmpeg stderr
                    if "ffmpeg version" in line:
                        ignore_line = True
                    if not ignore_line and line:
                        self._queue.put_nowait(line)

                ignore_line = False

            sleep(self.WAIT_TIME)

    def attach_filedescriptor(self, filedesc: IO[str] | None = None) -> None:
        """Attach a filedescriptor to the PipeReader."""
        self._filedescriptor = filedesc

    def join(self, timeout=None) -> None:
        self._running = False
        super().join(timeout)


class YoutubeDLDownloader:
    """Python class for downloading videos using youtube-dl & subprocess.

    Attributes:
        OK, ERROR, STOPPED, ALREADY, FILESIZE_ABORT, WARNING (int): Integers
            that describe the return code from the download() method. The
            larger the number the higher is the hierarchy of the code.
            Codes with smaller hierachy cannot overwrite codes with higher
            hierarchy.

    Args:
        youtubedl_path (str): Absolute path to youtube-dl binary.

        data_hook (Callable): Optional callback function to retrieve download
            process data.

        log_data (Callable): Optional callback function to write data to
            the log file.

    Warnings:
        The caller is responsible for calling the close() method after he has
        finished with the object in order for the object to be able to properly
        close down itself.

    Example:
        How to use YoutubeDLDownloader from a python script.

            from downloaders import YoutubeDLDownloader

            def data_hook(data):
                print(data)

            downloader = YoutubeDLDownloader('/usr/bin/youtube-dl', data_hook)

            downloader.download(<URL STRING>, ['-f', 'flv'])

    """

    OK = 0
    WARNING = 1
    ERROR = 2
    FILESIZE_ABORT = 3
    ALREADY = 4
    STOPPED = 5

    def __init__(
        self,
        youtubedl_path: str,
        data_hook: Callable[[dict[str, Any]], None] | None = None,
        log_data: Callable[[str], None] | None = None,
    ):
        self.youtubedl_path: str = youtubedl_path
        self.data_hook = data_hook
        self.log_data = log_data

        self._return_code: int = self.OK
        self._proc: subprocess.Popen | None = None

        self._stderr_queue: Queue = Queue()
        self._stderr_reader = PipeReader(self._stderr_queue)

    def download(self, url: str, options: list[str] | None = None) -> int:
        """Download url using given options.

        Args:
            url (str): URL string to download.
            options (list): Python list that contains youtube-dl options.

        Returns:
            An integer that shows the status of the download process.
            There are 6 different return codes.

            OK (0): The download process completed successfully.
            WARNING (1): A warning occured during the download process.
            ERROR (2): An error occured during the download process.
            FILESIZE_ABORT (3): The corresponding url video file was larger or
                smaller from the given filesize limit.
            ALREADY (4): The given url is already downloaded.
            STOPPED (5): The download process was stopped by the user.

        """
        self._return_code = self.OK

        cmd = self._get_cmd(url, options)
        self._create_process(cmd)

        if self._proc is not None:
            self._stderr_reader.attach_filedescriptor(self._proc.stderr)

        while self._proc_is_alive():
            stdout: str = ""
            if not self._proc.stdout.closed:
                with contextlib.suppress(ValueError):
                    stdout = self._proc.stdout.readline().rstrip()
            if stdout:
                data_dict = extract_data(stdout)
                self._extract_info(data_dict)
                self._hook_data(data_dict)

        # Read stderr after download process has been completed
        # We don't need to read stderr in real time
        while not self._stderr_queue.empty():
            stderr = str(self._stderr_queue.get_nowait()).rstrip()

            self._log(stderr)

            if self._is_warning(stderr):
                self._set_returncode(self.WARNING)

        if self._proc and self._proc.returncode > 0:
            proc_return_code = self._proc.returncode
            self._log(f"Child process exited with non-zero code: {proc_return_code}")
            self._set_returncode(self.ERROR)

        self._last_data_hook()

        return self._return_code

    def stop(self) -> None:
        """Stop the download process and set return code to STOPPED."""
        if not self._proc_is_alive():
            return

        self._proc.stdout.close()
        self._proc.stderr.close()

        with contextlib.suppress(ProcessLookupError):
            if IS_WINDOWS:
                # os.killpg is not available on Windows
                # See: https://bugs.python.org/issue5115
                self._proc.kill()

                # When we kill the child process on Windows the return code
                # gets set to 1, so we want to reset the return code back to 0
                # in order to avoid creating logging output in the download(...)
                # method
                self._proc.returncode = 0
            else:
                # TODO: Test in Unix os.killpg ?
                os.killpg(self._proc.pid, signal.SIGKILL)  # type: ignore
        self._set_returncode(self.STOPPED)

    def close(self) -> None:
        """Destructor like function for the object."""
        self._stderr_reader.join()

    def _set_returncode(self, code) -> None:
        """Set self._return_code only if the hierarchy of the given code is
        higher than the current self._return_code."""
        if code >= self._return_code:
            self._return_code = code

    @staticmethod
    def _is_warning(stderr: str) -> bool:
        warning_error = stderr.split(":")[0]
        warning_error = warning_error.strip()
        return warning_error in ["WARNING", "ERROR"]

    def _last_data_hook(self) -> None:
        """Set the last data information based on the return code."""
        data_dictionary: dict[str, str] = {
            "status": "",
            "speed": "",
            "eta": "",
        }

        if self._return_code == self.OK:
            data_dictionary["status"] = "Finished"
        elif self._return_code == self.ERROR:
            data_dictionary["status"] = "Error"
        elif self._return_code == self.WARNING:
            data_dictionary["status"] = "Warning"
        elif self._return_code == self.STOPPED:
            data_dictionary["status"] = "Stopped"
        elif self._return_code == self.ALREADY:
            data_dictionary["status"] = "Already Downloaded"
        else:
            data_dictionary["status"] = "Filesize Abort"

        self._hook_data(data_dictionary)

    def _extract_info(self, data: dict[str, Any]) -> None:
        """Extract informations about the download process from the given data.

        Args:
            data (dict): Python dictionary that contains different
                keys. The keys are not standar the dictionary can also be
                empty when there are no data to extract. See extract_data().

        """
        if "status" in data:
            if data["status"] == "Already Downloaded":
                # Set self._return_code to already downloaded
                # and trash that key
                self._set_returncode(self.ALREADY)
                data["status"] = None

            if data["status"] == "Filesize Abort":
                # Set self._return_code to filesize abort
                # and trash that key
                self._set_returncode(self.FILESIZE_ABORT)
                data["status"] = None

    def _log(self, data: str) -> None:
        """Log data using the callback function."""
        if self.log_data is not None:
            self.log_data(data)

    def _hook_data(self, data: dict[str, Any]):
        """Pass data back to the caller."""
        if self.data_hook is not None:
            self.data_hook(data)

    def _proc_is_alive(self) -> bool:
        """Returns True if self._proc is alive else False."""
        return False if self._proc is None else self._proc.poll() is None

    def _get_cmd(self, url: str, options: list[str] | None = None) -> list[str]:
        """Build the subprocess command.

        Args:
            url (str): URL string to download.
            options (list): Python list that contains youtube-dl options.

        Returns:
            Python list that contains the command to execute.

        """
        cmd_list: list[str] = [self.youtubedl_path]

        if options:
            cmd_list.extend(options)

        cmd_list.append(url)
        return cmd_list

    def _create_process(self, cmd: list[str]) -> None:
        """Create new subprocess.

        Args:
            cmd (list): Python list that contains the command to execute.

        """
        info = None

        kwargs = dict(
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            encoding=get_encoding(),
            creationflags=0,
        )

        if os.name == "nt":
            # Hide subprocess window
            info = subprocess.STARTUPINFO()  # type: ignore[attr-defined]
            info.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # type: ignore[attr-defined]
            info.wShowWindow = subprocess.SW_HIDE  # type: ignore[attr-defined]

            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP  # type: ignore[attr-defined]
        else:
            kwargs["start_new_session"] = True

        try:
            self._proc = subprocess.Popen(cmd, startupinfo=info, **kwargs)  # type: ignore
        except (ValueError, OSError) as error:
            self._log(f"Failed to start process: {cmd}")
            self._log(str(error))


def extract_filename(input_data: str) -> tuple[str, str, str]:
    """Extract the component of the filename

    Args:
        input_data (str): Filename with extension

    Returns:
        Python tuple with path, filename and extension

    """
    _filename = Path(input_data.strip('"'))
    path: str = str(_filename.parent) if str(_filename.parent) != "." else ""
    filename: str = _filename.stem
    extension: str = _filename.suffix

    return path, filename, extension


def extract_data(stdout: str) -> dict[str, str]:
    """Extract data from youtube-dl stdout.

    Args:
        stdout (str): String that contains the youtube-dl stdout.

    Returns:
        Python dictionary. The returned dictionary can be empty if there are
        no data to extract else it may contain one or more of the
        following keys:

        'status'         : Contains the status of the download process.
        'path'           : Destination path.
        'extension'      : The file extension.
        'filename'       : The filename without the extension.
        'percent'        : The percentage of the video being downloaded.
        'eta'            : Estimated time for the completion of the download process.
        'speed'          : Download speed.
        'filesize'       : The size of the video file being downloaded.
        'playlist_index' : The playlist index of the current video file being downloaded.
        'playlist_size'  : The number of videos in the playlist.

    """
    # REFACTOR
    # noinspection PyShadowingNames

    data_dictionary: dict[str, str] = {}

    if not stdout:
        return data_dictionary

    # We want to keep the spaces in order to extract filenames with
    # multiple whitespaces correctly.
    # Fix output of yt-dlp compatible with youtube-dl
    stdout_list: list[str] = re.sub(r"~\s+", "~", stdout, count=1).split()

    stdout_list[0] = stdout_list[0].lstrip("\r")

    if stdout_list[0] == "[download]":
        data_dictionary["status"] = "Downloading"

        # Get path, filename & extension
        if stdout_list[1] == "Destination:":
            path, filename, extension = extract_filename(" ".join(stdout_list[2:]))

            data_dictionary["path"] = path
            data_dictionary["filename"] = filename
            data_dictionary["extension"] = extension

        # Get progress info
        if "%" in stdout_list[1]:
            if stdout_list[1] == "100%":
                data_dictionary["speed"] = ""
                data_dictionary["eta"] = ""
                data_dictionary["percent"] = "100%"
                data_dictionary["filesize"] = stdout_list[3]
            else:
                data_dictionary["percent"] = stdout_list[1]
                data_dictionary["filesize"] = stdout_list[3]
                data_dictionary["speed"] = stdout_list[5]
                data_dictionary["eta"] = stdout_list[7]

        # Get playlist info
        if stdout_list[1] == "Downloading" and stdout_list[2] == "video":
            data_dictionary["playlist_index"] = stdout_list[3]
            data_dictionary["playlist_size"] = stdout_list[5]

        # Remove the 'and merged' part from stdout when using ffmpeg to merge the formats
        if stdout_list[-3] == "downloaded" and stdout_list[-1] == "merged":
            stdout_list = stdout_list[:-2]
            data_dictionary["percent"] = "100%"

        # Get file already downloaded status
        if stdout_list[-1] == "downloaded":
            data_dictionary["status"] = "Already Downloaded"
            path, filename, extension = extract_filename(" ".join(stdout_list[1:-4]))

            data_dictionary["path"] = path
            data_dictionary["filename"] = filename
            data_dictionary["extension"] = extension

        # Get filesize abort status
        if stdout_list[-1] == "Aborting.":
            data_dictionary["status"] = "Filesize Abort"

    elif stdout_list[0] == "[hlsnative]":
        # native hls extractor
        # see: https://github.com/rg3/youtube-dl/blob/master/youtube_dl/downloader/hls.py#L54
        data_dictionary["status"] = "Downloading"

        if len(stdout_list) == 7:
            segment_no = float(stdout_list[6])
            current_segment = float(stdout_list[4])

            # Get the percentage
            percent = f"{current_segment / segment_no * 100:.1f}%"
            data_dictionary["percent"] = percent

    elif stdout_list[0] == "[ffmpeg]":
        data_dictionary["status"] = "Post Processing"

        # Get final extension after merging process
        if stdout_list[1] == "Merging":
            path, filename, extension = extract_filename(" ".join(stdout_list[4:]))

            data_dictionary["path"] = path
            data_dictionary["filename"] = filename
            data_dictionary["extension"] = extension

        # Get final extension ffmpeg post process simple (not file merge)
        if stdout_list[1] == "Destination:":
            path, filename, extension = extract_filename(" ".join(stdout_list[2:]))

            data_dictionary["path"] = path
            data_dictionary["filename"] = filename
            data_dictionary["extension"] = extension

        # Get final extension after recoding process
        if stdout_list[1] == "Converting":
            path, filename, extension = extract_filename(" ".join(stdout_list[8:]))

            data_dictionary["path"] = path
            data_dictionary["filename"] = filename
            data_dictionary["extension"] = extension

    elif stdout_list[0][0] == "[" and stdout_list[0] != "[debug]":
        data_dictionary["status"] = "Pre Processing"

    return data_dictionary



#!/usr/bin/env python

"""Contains test cases for the downloadmanager.py module."""


import sys
import unittest
from pathlib import Path
from unittest import mock

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.downloadmanager import DownloadItem, DownloadList, DownloadManager


class TestDownloadManager(unittest.TestCase):
    @mock.patch("youtube_dl_gui.downloadmanager.DownloadManager", autospec=True)
    def test_init_check_sig(self, mock_dwlmng):
        dwn_manager = mock_dwlmng(parent=None, download_list=[], opt_manager=None)
        dwn_manager._running = True
        self.assertIsInstance(dwn_manager, DownloadManager)
        self.assertTrue(dwn_manager._running)

    @mock.patch("youtube_dl_gui.downloadmanager.Worker")
    @mock.patch("youtube_dl_gui.mainframe.MainFrame")
    @mock.patch("youtube_dl_gui.optionsmanager.OptionsManager")
    def test_downloadmanager(self, mock_opt_manager, mock_mainframe, mock_worker):
        config_path = "/home/user/.config"
        dwl_list = DownloadList(
            [
                DownloadItem("url1", ["-v", "-F"]),
                DownloadItem("url2", ["-v", "-F"]),
                DownloadItem("url3", ["-v", "-F"]),
            ]
        )
        opt_manager = mock_opt_manager(config_path)
        opt_manager.options = {
            "youtubedl_path": config_path,
            "workers_number": 3,
            "disable_update": True,
        }
        parent = mock_mainframe(opt_manager)
        mock_worker.available.return_value = True
        # End thread for this test like daemon
        # TODO: Finished join the thread
        dwn_manager = DownloadManager(parent, dwl_list, opt_manager, daemon=True)

        self.assertTrue(dwn_manager._running)
        self.assertEqual(dwn_manager.name, "DownloadManager")
        self.assertEqual(len(dwn_manager._workers), 3)
        dwn_manager.stop_downloads()
        self.assertFalse(dwn_manager._running)
        self.assertTrue(dwn_manager._jobs_done())


def main():
    unittest.main()


if __name__ == "__main__":
    main()



from __future__ import annotations

import wx

DARK_BACKGROUND_COLOUR = wx.Colour(29, 31, 33, 255)
DARK_FOREGROUND_COLOUR = wx.Colour(197, 200, 198, 255)
DARK_BACKGROUND_COLOUR_BUTTON = wx.Colour(48, 48, 49, 255)
DARK_LIGHTGREY_COLOUR = wx.Colour(240, 240, 240, 255)


def get_widgets(parent: wx.Window | wx.Panel) -> list[wx.Window]:
    """Return a list of all the child widgets"""

    items: list[wx.Window] = [parent]

    for item in parent.GetChildren():
        items.append(item)

        if hasattr(item, "GetChildren"):
            for child in item.GetChildren():
                items.append(child)

    return items


def dark_row_formatter(listctrl: wx.ListCtrl, dark: bool = False) -> None:
    """Toggles the row in a ListCtrl"""

    items = [listctrl.GetItem(i) for i in range(listctrl.GetItemCount())]

    for index, item in enumerate(items):
        if dark:
            if index % 2:
                item.SetBackgroundColor(DARK_BACKGROUND_COLOUR)
            else:
                item.SetBackgroundColor(DARK_LIGHTGREY_COLOUR)

        listctrl.SetItem(item)


def dark_mode(parent: wx.Window | wx.Panel, _dark_mode: bool = False) -> None:
    """Toggles dark mode"""

    widgets: list[wx.Window] = get_widgets(parent)
    # panel = widgets[0]

    for widget in widgets:
        if _dark_mode and not isinstance(widget, (wx.TextCtrl, wx.StaticLine)):
            widget.SetBackgroundColour(DARK_BACKGROUND_COLOUR)
            widget.SetForegroundColour(DARK_FOREGROUND_COLOUR)

            if isinstance(widget, wx.ListCtrl):
                dark_row_formatter(widget, dark=True)
            elif isinstance(widget, wx.Button):
                widget.SetBackgroundColour(DARK_BACKGROUND_COLOUR_BUTTON)

    parent.Refresh()



"""yt-dlg module to update youtube-dl binary.

Attributes:
    UPDATE_PUB_TOPIC (str): wxPublisher subscription topic of the
        UpdateThread thread.

"""
from __future__ import annotations

import json
import os
import stat
from pathlib import Path
from threading import Thread
from typing import TYPE_CHECKING, Any
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

import wx

# noinspection PyPep8Naming
from pubsub import pub as Publisher

from .utils import IS_WINDOWS, YOUTUBEDL_BIN, YTDLP_BIN, check_path

UPDATE_PUB_TOPIC = "update"

if TYPE_CHECKING:
    from .optionsmanager import OptionsManager


class UpdateThread(Thread):
    """Python Thread that downloads youtube-dl binary.

    Attributes:
        LATEST_YOUTUBE_DL (str): URL with the latest youtube-dl binary.

        GITHUB_API (str): URL GitHub API v3

        LATEST_YOUTUBE_DL_API (str): See `LATEST_YOUTUBE_DL` attribute

        DOWNLOAD_TIMEOUT (int): Download timeout in seconds.

    Args:
        opt_manager (optionsmanager.OptionsManager): Options manager

        quiet (bool): If True UpdateThread won't send the finish signal
            back to the caller. Finish signal can be used to make sure that
            the UpdateThread has been completed in an asynchronous way.

        daemon (bool): If Thread is daemonic

    """

    LATEST_YOUTUBE_DL: str = "https://yt-dl.org/latest/"
    GITHUB_API: str = "https://api.github.com/"
    LATEST_YOUTUBE_DL_API: str = (
        GITHUB_API + "repos/ytdl-org/youtube-dl/releases/latest"
    )
    DOWNLOAD_TIMEOUT: int = 10

    def __init__(
        self,
        opt_manager: OptionsManager,
        quiet: bool = False,
        daemon: bool = False,
    ):
        super().__init__()
        self.opt_manager = opt_manager
        self.download_path: str = opt_manager.options.get("youtubedl_path", ".")
        self.cli_backend: str = opt_manager.options.get("cli_backend", YOUTUBEDL_BIN)
        self.quiet: bool = quiet

        if self.cli_backend == YTDLP_BIN:
            self.LATEST_YOUTUBE_DL = "https://github.com/yt-dlp/yt-dlp/releases/"
            self.LATEST_YOUTUBE_DL_API = (
                f"{self.GITHUB_API}repos/yt-dlp/yt-dlp/releases/latest"
            )

        self.name = "UpdateManager"
        self.daemon = daemon
        self.start()

    def get_latest_sourcefile(self) -> str:
        """Get the URL file name of the latest asset"""
        source_file: str = self.GITHUB_API
        try:
            stream = urlopen(self.LATEST_YOUTUBE_DL_API, timeout=self.DOWNLOAD_TIMEOUT)

            latest_json: dict[str, Any] = json.load(stream)
            latest_assets: list[dict[str, Any]] = latest_json["assets"]

            for asset in latest_assets:
                if asset["name"] == self.cli_backend:
                    source_file = asset["browser_download_url"]
                    break
        except (HTTPError, URLError, json.JSONDecodeError) as error:
            self._talk_to_gui("error", str(error))

        return source_file

    def run(self) -> None:
        self._talk_to_gui("download")

        source_file: str = self.get_latest_sourcefile()
        destination_file: str = str(Path(self.download_path) / Path(self.cli_backend))

        check_path(self.download_path)

        try:
            stream = urlopen(source_file, timeout=self.DOWNLOAD_TIMEOUT)

            with open(destination_file, "wb") as dest_file:
                dest_file.write(stream.read())

            # Have to set the executable flag on linux
            if not IS_WINDOWS:
                mode = os.stat(destination_file).st_mode
                mode |= mode | stat.S_IEXEC
                os.chmod(destination_file, mode)

            self._talk_to_gui("correct")
        except (HTTPError, URLError, OSError) as error:
            self._talk_to_gui("error", str(error))

        if not self.quiet:
            self._talk_to_gui("finish")

    @staticmethod
    def _talk_to_gui(signal: str, data: str | None = None) -> None:
        """Communicate with the GUI using wxCallAfter and wxPublisher.

        Args:
            signal (str): Unique signal string that informs the GUI for the
                update process.

            data (str): Can be any string data to pass along with the
                given signal. Default is None.

        Note:
            UpdateThread supports 4 signals.
                1) download: The update process started
                2) correct: The update process completed successfully
                3) error: An error occured while downloading youtube-dl binary
                4) finish: The update thread is ready to join

        """
        if wx.GetApp() is not None:
            wx.CallAfter(
                Publisher.sendMessage, UPDATE_PUB_TOPIC, signal=signal, data=data
            )



"""Contains test cases for the DownloadList object."""

import os.path
import sys
import unittest
from unittest import mock

PATH = os.path.realpath(os.path.abspath(__file__))
sys.path.insert(0, os.path.dirname(os.path.dirname(PATH)))

from youtube_dl_gui.downloadmanager import DownloadList, synchronized


class TestInit(unittest.TestCase):

    """Test case for the DownloadList init."""

    def test_init(self):
        mocks = [mock.Mock(object_id=0), mock.Mock(object_id=1)]

        dlist = DownloadList(mocks)  # type: ignore
        self.assertEqual(dlist._items_list, [0, 1])
        self.assertEqual(dlist._items_dict, {0: mocks[0], 1: mocks[1]})

    def test_init_empty(self):
        dlist = DownloadList()
        self.assertEqual(dlist._items_list, [])
        self.assertEqual(dlist._items_dict, {})

    def test_init_invalid_args(self):
        self.assertRaises(AssertionError, DownloadList, {})
        self.assertRaises(AssertionError, DownloadList, ())
        self.assertRaises(AssertionError, DownloadList, False)


class TestInsert(unittest.TestCase):

    """Test case for the DownloadList insert method."""

    def test_insert(self):
        mock_ditem = mock.Mock(object_id=0)

        dlist = DownloadList()
        dlist.insert(mock_ditem)

        self.assertEqual(dlist._items_list, [0])
        self.assertEqual(dlist._items_dict, {0: mock_ditem})


class TestRemove(unittest.TestCase):

    """Test case for the DownloadList remove method."""

    def setUp(self):
        self.mocks = [
            mock.Mock(object_id=0),
            mock.Mock(object_id=1),
            mock.Mock(object_id=2),
        ]
        self.dlist = DownloadList(self.mocks)  # type: ignore

    def test_remove(self):
        self.assertTrue(self.dlist.remove(1))

        self.assertEqual(self.dlist._items_list, [0, 2])
        self.assertEqual(self.dlist._items_dict, {0: self.mocks[0], 2: self.mocks[2]})

    def test_remove_not_exist(self):
        self.assertRaises(KeyError, self.dlist.remove, 3)

    def test_remove_active(self):
        self.mocks[1].stage = "Active"

        self.assertFalse(self.dlist.remove(1))
        self.assertEqual(self.dlist._items_list, [0, 1, 2])
        self.assertEqual(
            self.dlist._items_dict,
            {0: self.mocks[0], 1: self.mocks[1], 2: self.mocks[2]},
        )


class TestFetchNext(unittest.TestCase):

    """Test case for the DownloadList fetch_next method."""

    def test_fetch_next(self):
        items_count = 3

        mocks = [mock.Mock(object_id=i, stage="Queued") for i in range(items_count)]

        dlist = DownloadList(mocks)  # type: ignore

        for i in range(items_count):
            self.assertEqual(dlist.fetch_next(), mocks[i])
            mocks[i].stage = "Active"

        self.assertIsNone(dlist.fetch_next())

        for i in range(items_count):
            mocks[i].stage = "Completed"

        self.assertIsNone(dlist.fetch_next())

        mocks[1].stage = "Queued"  # Re-queue item
        self.assertEqual(dlist.fetch_next(), mocks[1])

    def test_fetch_next_empty_list(self):
        dlist = DownloadList()
        self.assertIsNone(dlist.fetch_next())


class TestMoveUp(unittest.TestCase):

    """Test case for the DownloadList move_up method."""

    def setUp(self):
        mocks = [mock.Mock(object_id=i, stage="Queued") for i in range(3)]
        self.dlist = DownloadList(mocks)  # type: ignore

    def test_move_up(self):
        self.assertTrue(self.dlist.move_up(1))
        self.assertEqual(self.dlist._items_list, [1, 0, 2])

    def test_move_up_already_on_top(self):
        self.assertFalse(self.dlist.move_up(0))
        self.assertEqual(self.dlist._items_list, [0, 1, 2])

    def test_move_up_not_exist(self):
        self.assertRaises(ValueError, self.dlist.move_up, 666)


class TestMoveDown(unittest.TestCase):

    """Test case for the DownloadList move_down method."""

    def setUp(self):
        mocks = [mock.Mock(object_id=i, stage="Queued") for i in range(3)]
        self.dlist = DownloadList(mocks)  # type: ignore

    def test_move_down(self):
        self.assertTrue(self.dlist.move_down(1))
        self.assertEqual(self.dlist._items_list, [0, 2, 1])

    def test_move_down_already_on_bottom(self):
        self.assertFalse(self.dlist.move_down(2))
        self.assertEqual(self.dlist._items_list, [0, 1, 2])

    def test_move_down_not_exist(self):
        self.assertRaises(ValueError, self.dlist.move_down, 666)


class TestGetItem(unittest.TestCase):

    """Test case for the DownloadList get_item method."""

    def test_get_item(self):
        mocks = [mock.Mock(object_id=i) for i in range(3)]
        dlist = DownloadList(mocks)  # type: ignore

        self.assertEqual(dlist.get_item(0), mocks[0])
        self.assertEqual(dlist.get_item(2), mocks[2])

    def test_get_item_not_exist(self):
        dlist = DownloadList()
        self.assertIsNone(dlist.get_item(0))


class TestGetLength(unittest.TestCase):

    """Test case for the DownloadList __len__ method."""

    def test_get_length(self):
        dlist = DownloadList([mock.Mock(), mock.Mock()])
        self.assertEqual(len(dlist), 2)

    def test_get_length_empty_list(self):
        dlist = DownloadList()
        self.assertEqual(len(dlist), 0)


class TestHasItem(unittest.TestCase):

    """Test case for the DownloadList has_item method."""

    def setUp(self):
        mock_ditem = mock.Mock(object_id=1337)
        self.dlist = DownloadList([mock_ditem])

    def test_has_item_true(self):
        self.assertTrue(self.dlist.has_item(1337))

    def test_has_item_false(self):
        self.assertFalse(self.dlist.has_item(1000))


class TestGetItems(unittest.TestCase):

    """Test case for the DownloadList get_items method."""

    def test_get_items(self):
        mocks = [mock.Mock() for _ in range(3)]
        dlist = DownloadList(mocks)  # type: ignore

        self.assertEqual(dlist.get_items(), mocks)

    def test_get_items_empty_list(self):
        dlist = DownloadList()
        self.assertEqual(dlist.get_items(), [])


class TestClear(unittest.TestCase):

    """Test case for the DownloadList clear method."""

    def test_clear(self):
        dlist = DownloadList([mock.Mock() for _ in range(3)])

        self.assertEqual(len(dlist), 3)
        dlist.clear()
        self.assertEqual(len(dlist), 0)


class TestChangeStage(unittest.TestCase):

    """Test case for the DownloadList change_stage method."""

    def setUp(self):
        self.mocks = [mock.Mock(object_id=i, stage="Queued") for i in range(3)]
        self.dlist = DownloadList(self.mocks)  # type: ignore

    def test_change_stage(self):
        self.dlist.change_stage(0, "Active")
        self.assertEqual(self.mocks[0].stage, "Active")

    def test_change_stage_id_not_exist(self):
        self.assertRaises(KeyError, self.dlist.change_stage, 3, "Active")


class TestIndex(unittest.TestCase):

    """Test case for the DownloadList index method."""

    def setUp(self):
        self.mocks = [mock.Mock(object_id=i) for i in range(3)]
        self.dlist = DownloadList(self.mocks)  # type: ignore

    def test_index(self):
        self.assertEqual(self.dlist.index(2), 2)

    def test_index_not_exist(self):
        self.assertEqual(self.dlist.index(3), -1)


class TestSynchronizeDecorator(unittest.TestCase):
    def test_synchronize(self):
        mock_func = mock.Mock()
        mock_lock = mock.Mock()

        decorated_func = synchronized(mock_lock)(mock_func)

        self.assertEqual(decorated_func(1, a=2), mock_func.return_value)

        mock_func.assert_called_once_with(1, a=2)
        mock_lock.acquire.assert_called_once()
        mock_lock.release.assert_called_once()


def main():
    unittest.main()


if __name__ == "__main__":
    main()



"""Contains test cases for the MessageDialog in widgets.py module."""


import sys
import unittest
from pathlib import Path

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))


from tests.wtc import WidgetTestCase

from youtube_dl_gui.widgets import MessageDialog


class TestMessageDialog(WidgetTestCase):
    """Test cases for the MessageDialog widget."""

    def test_init_dark(self):
        _dark_mode = True
        msg_dlg = MessageDialog(self.frame, "Test MessageDialog", "Test", _dark_mode)
        self.assertTrue(msg_dlg._dark_mode)
        self.assertEqual(len(msg_dlg.buttons), 2)


if __name__ == "__main__":
    unittest.main()



"""yt-dlg module to handle settings. """
from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any

from .formats import FORMATS, OUTPUT_FORMATS
from .utils import YTDLP_BIN, check_path, decode_tuple, encode_tuple, get_default_lang


class OptionsManager:
    # noinspection PyUnresolvedReferences
    """Handles yt-dlg options.

    This class is responsible for storing and retrieving the options.

    Attributes:
        SETTINGS_FILENAME (str): Filename of the settings file.
        SENSITIVE_KEYS (tuple): Contains the keys that we don't want
            to store on the settings file. (SECURITY ISSUES).

    Args:
        config_path (str): Absolute path where OptionsManager
            should store the settings file.

    Note:
        See load_default() method for available options.

    Example:
        Access the options using the 'options' variable.

        opt_manager = OptionsManager('.')
        opt_manager.options['save_path'] = '~/Downloads'

    """

    SETTINGS_FILENAME = "settings.json"
    SENSITIVE_KEYS = ("sudo_password", "password", "video_password")
    MAIN_WIN_SIZE = (740, 490)
    OPTS_WIN_SIZE = (640, 490)

    def __init__(self, config_path: str):
        self.config_path: str = config_path
        self.settings_file: str = str(Path(config_path) / Path(self.SETTINGS_FILENAME))
        self.options: dict[str, Any] = {}
        self.load_default()
        self.load_from_file()

    def load_default(self) -> None:
        """Load the default options.

        Note:
            This method is automatically called by the constructor.

        Options Description:

            save_path (str): Path where youtube-dl should store the
                downloaded file. Default is $HOME.

            video_format (str): Video format to download.
                When this options is set to '0' youtube-dl will choose
                the best video format available for the given URL.

            second_video_format (str): Video format to mix with the first
                one (-f 18+17).

            to_audio (bool): If True youtube-dl will post process the
                video file.

            keep_video (bool): If True youtube-dl will keep the video file
                after post processing it.

            audio_format (str): Audio format of the post processed file.
                Available values are "mp3", "wav", "aac", "m4a", "vorbis",
                "opus" & "flac".

            audio_quality (str): Audio quality of the post processed file.
                Available values are "9", "5", "0". The lowest the value the
                better the quality.

            restrict_filenames (bool): If True youtube-dl will restrict
                the downloaded file filename to ASCII characters only.

            output_format (str): This option sets the downloaded file
                output template. See formats.OUTPUT_FORMATS for more info.

            output_template (str): Can be any output template supported
                by youtube-dl.

            playlist_start (int): Playlist index to start downloading.

            playlist_end (int): Playlist index to stop downloading.

            max_downloads (int): Maximum number of video files to download
                from the given playlist.

            min_filesize (float): Minimum file size of the video file.
                If the video file is smaller than the given size then
                youtube-dl will abort the download process.

            max_filesize (float): Maximum file size of the video file.
                If the video file is larger than the given size then
                youtube-dl will abort the download process.

            min_filesize_unit (str): Minimum file size unit.
                Available values: '', 'k', 'm', 'g', 'y', 'p', 'e', 'z', 'y'.

            max_filesize_unit (str): Maximum file size unit.
                See 'min_filesize_unit' option for available values.

            write_subs (bool): If True youtube-dl will try to download
                the subtitles file for the given URL.

            write_all_subs (bool): If True youtube-dl will try to download
                all the available subtitles files for the given URL.

            write_auto_subs (bool): If True youtube-dl will try to download
                the automatic subtitles file for the given URL.

            embed_subs (bool): If True youtube-dl will merge the subtitles
                file with the video. (ONLY mp4 files).

            subs_lang (str): Language of the subtitles file to download.
                Needs 'write_subs' option.

            ignore_errors (bool): If True youtube-dl will ignore the errors
                and continue the download process.

            open_dl_dir (bool): If True yt-dlg will open the
                destination folder after download process has been completed.

            write_description (bool): If True youtube-dl will write video
                description to a .description file.

            write_info (bool): If True youtube-dl will write video
                metadata to a .info.json file.

            write_thumbnail (bool): If True youtube-dl will write
                thumbnail image to disk.

            retries (int): Number of youtube-dl retries.

            user_agent (str): Specify a custom user agent for youtube-dl.

            referer (str): Specify a custom referer to use if the video
                access is restricted to one domain.

            proxy (str): Use the specified HTTP/HTTPS proxy.

            shutdown (bool): If True yt-dlg will turn the computer
                off after the download process has been completed.

            sudo_password (str): SUDO password for the shutdown process if
                the user does not have elevated privileges.

            username (str): Username to login with.

            password (str): Password to login with.

            video_password (str): Video password for the given URL.

            cli_backend (str): CLI backen used to download.
                Currently youtube-dl, yt-dlp
                Default youtube-dl

            youtubedl_path (str): Absolute path to the youtube-dl binary.
                Default is the self.config_path. You can change this option
                to point on /usr/local/bin etc.. if you want to use the
                youtube-dl binary on your system. This is also the directory
                where yt-dlg will auto download the youtube-dl if not
                exists so you should make sure you have write access if you
                want to update the youtube-dl binary from within yt-dlg.

            cmd_args (str): String that contains extra youtube-dl options
                seperated by spaces.

            enable_log (bool): If True yt-dlg will enable
                the LogManager. See app module.

            log_time (bool): See logmanager.LogManager add_time attribute.

            workers_number (int): Number of download workers that download manager
                will spawn. Must be greater than zero.

            locale_name (str): Locale name (e.g. ru_RU).

            main_win_size (tuple): Main window size (width, height).
                If window becomes to small the program will reset its size.
                See _settings_are_valid method MIN_FRAME_SIZE.

            opts_win_size (tuple): Options window size (width, height).
                If window becomes to small the program will reset its size.
                See _settings_are_valid method MIN_FRAME_SIZE.

            save_path_dirs (list): List that contains temporary save paths.

            selected_video_formats (list): List that contains the selected
                video formats to display on the main window.

            selected_audio_formats (list): List that contains the selected
                audio formats to display on the main window.

            selected_format (str): Current format selected on the main window.

            youtube_dl_debug (bool): When True will pass '-v' flag to youtube-dl.

            ignore_config (bool): When True will ignore youtube-dl config file options.

            confirm_exit (bool): When True create popup to confirm exiting youtube-dl-gui.

            native_hls (bool): When True youtube-dl will use the native HLS implementation.

            # TODO:
            ffmpeg_hls (bool): (--hls-prefer-ffmpeg) Use ffmpeg insted of the native HLS
                downloader. (IMPLEMENT with radio buttons)
                EX: --hls-prefer-ffmpeg \
                    --external-downloader ffmpeg --external-downloader-args "-ss 234 -to 621"
            # TODO:
            external_downloader (str): Use the specified external downloader.
                Currently supports aria2c, avconv, axel, curl, ffmpeg, httpie, wget
                Default ffmpeg
            # TODO:
            external_downloader_args (list): List of string give to arguments
                to the external downloader.

            show_completion_popup (bool): When True youtube-dl-gui will create a popup
                to inform the user for the download completion.

            confirm_deletion (bool): When True ask user before item removal.

            nomtime (bool): When True will not use the Last-modified header to
                set the file modification time.

            embed_thumbnail (bool): When True will embed the thumbnail in
                the audio file as cover art.

            add_metadata (bool): When True will write metadata to file.

            disable_update (bool): When True the update process will be disabled.

        """
        # REFACTOR Remove old options & check options validation
        self.options = {
            "save_path": str(Path().home()),
            "save_path_dirs": [
                str(Path().home()),
                str(Path().home() / Path("Downloads")),
                str(Path().home() / Path("Desktop")),
                str(Path().home() / Path("Videos")),
                str(Path().home() / Path("Music")),
            ],
            "video_format": "0",
            "second_video_format": "0",
            "to_audio": False,
            "keep_video": False,
            "audio_format": "",
            "audio_quality": "5",
            "restrict_filenames": False,
            "dark_mode": False,
            "output_format": "1",
            "output_template": str(Path("%(uploader)s") / Path("%(title)s.%(ext)s")),
            "playlist_start": 1,
            "playlist_end": 0,
            "max_downloads": 0,
            "min_filesize": 0,
            "max_filesize": 0,
            "min_filesize_unit": "",
            "max_filesize_unit": "",
            "write_subs": False,
            "write_all_subs": False,
            "write_auto_subs": False,
            "embed_subs": False,
            "subs_lang": "en",
            "ignore_errors": True,
            "open_dl_dir": False,
            "write_description": False,
            "write_info": False,
            "write_thumbnail": False,
            "retries": 10,
            "user_agent": "",
            "referer": "",
            "proxy": "",
            "shutdown": False,
            "sudo_password": "",
            "username": "",
            "password": "",
            "video_password": "",
            "cli_backend": YTDLP_BIN,
            "youtubedl_path": self.config_path,
            "cmd_args": "",
            "enable_log": True,
            "log_time": True,
            "workers_number": 3,
            "locale_name": get_default_lang(),
            "main_win_size": self.MAIN_WIN_SIZE,
            "opts_win_size": self.OPTS_WIN_SIZE,
            "selected_video_formats": ["webm", "mp4"],
            "selected_audio_formats": ["mp3", "m4a", "vorbis"],
            "selected_format": "0",
            "youtube_dl_debug": False,
            "ignore_config": True,
            "confirm_exit": True,
            "native_hls": True,
            "show_completion_popup": True,
            "confirm_deletion": True,
            "nomtime": False,
            "embed_thumbnail": False,
            "add_metadata": False,
            "disable_update": False,
        }

        # Set the youtubedl_path again if the disable_update option is set
        new_path: str = "/usr/bin"

        if (
            self.options["disable_update"]
            and os.name != "nt"
            and Path(new_path).exists()
        ):
            self.options["youtubedl_path"] = new_path

    def load_from_file(self) -> None:
        """Load options from settings file."""
        settings_path: Path = Path(self.settings_file)

        if not settings_path.exists():
            return

        with open(settings_path) as settings_file:
            try:
                options: dict[str, Any] = json.load(settings_file)

                if self._settings_are_valid(options):
                    self.options = options
            except json.JSONDecodeError:
                self.load_default()

    def save_to_file(self) -> None:
        """Save options to settings file."""
        check_path(self.config_path)

        with open(self.settings_file, "w") as settings_file:
            options = self._get_options()
            json.dump(options, settings_file, indent=4, separators=(",", ": "))

    # noinspection PyPep8Naming
    def _settings_are_valid(self, settings_dictionary: dict[str, Any]) -> bool:
        """Check settings.json dictionary.

        Args:
            settings_dictionary (dict): Options dictionary loaded
                from the settings file. See load_from_file() method.

        Returns:
            True if settings.json dictionary is valid, else False.

        """
        VALID_VIDEO_FORMAT = (
            "0",
            "17",
            "36",
            "5",
            "34",
            "35",
            "43",
            "44",
            "45",
            "46",
            "18",
            "22",
            "37",
            "38",
            "160",
            "133",
            "134",
            "135",
            "136",
            "137",
            "264",
            "138",
            "242",
            "243",
            "244",
            "247",
            "248",
            "271",
            "272",
            "82",
            "83",
            "84",
            "85",
            "100",
            "101",
            "102",
            "139",
            "140",
            "141",
            "171",
            "172",
        )

        VALID_AUDIO_FORMAT = ("mp3", "wav", "aac", "m4a", "vorbis", "opus", "flac", "")

        VALID_AUDIO_QUALITY = ("0", "5", "9")

        VALID_FILESIZE_UNIT = ("", "k", "m", "g", "t", "p", "e", "z", "y")

        VALID_SUB_LANGUAGE = (
            "en",
            "el",
            "pt",
            "fr",
            "it",
            "ru",
            "es",
            "de",
            "he",
            "sv",
            "tr",
        )

        MIN_FRAME_SIZE = 100

        # Decode string formatted tuples back to normal tuples
        settings_dictionary["main_win_size"] = decode_tuple(
            settings_dictionary["main_win_size"]
        )
        settings_dictionary["opts_win_size"] = decode_tuple(
            settings_dictionary["opts_win_size"]
        )

        for key in self.options:
            if key not in settings_dictionary:
                return False

            if not isinstance(self.options[key], type(settings_dictionary[key])):
                return False

        # Check if each key has a valid value
        rules_dict = {
            "video_format": FORMATS.keys(),
            "second_video_format": VALID_VIDEO_FORMAT,
            "audio_format": VALID_AUDIO_FORMAT,
            "audio_quality": VALID_AUDIO_QUALITY,
            "output_format": OUTPUT_FORMATS.keys(),
            "min_filesize_unit": VALID_FILESIZE_UNIT,
            "max_filesize_unit": VALID_FILESIZE_UNIT,
            "subs_lang": VALID_SUB_LANGUAGE,
        }

        for key, valid_list in rules_dict.items():
            if settings_dictionary[key] not in valid_list:
                return False

        # Check workers number value
        if settings_dictionary["workers_number"] < 1:
            return False

        # Check main-options frame size
        for size in settings_dictionary["main_win_size"]:
            if size < MIN_FRAME_SIZE:
                return False

        return all(
            size >= MIN_FRAME_SIZE for size in settings_dictionary["opts_win_size"]
        )

    def _get_options(self) -> dict[str, Any]:
        """Return options dictionary without SENSITIVE_KEYS."""
        temp_options = self.options.copy()

        for key in self.SENSITIVE_KEYS:
            temp_options[key] = ""

        # Encode normal tuples to string formatted tuples
        temp_options["main_win_size"] = encode_tuple(temp_options["main_win_size"])
        temp_options["opts_win_size"] = encode_tuple(temp_options["opts_win_size"])

        return temp_options



"""Contains test cases for the ClipDialog in widgets.py module."""

import sys
import unittest
from pathlib import Path
from typing import TYPE_CHECKING, cast

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))


from tests.wtc import WidgetTestCase

from youtube_dl_gui.downloadmanager import DownloadItem
from youtube_dl_gui.widgets import ClipDialog

if TYPE_CHECKING:
    from youtube_dl_gui.mainframe import MainFrame


class TestClipDialog(WidgetTestCase):
    """Test cases for the ClipDialog widget."""

    def test_get_timespans(self):
        ditem = DownloadItem("url", ["-f", "flv"])
        self.dlg = ClipDialog(cast("MainFrame", self.frame), ditem)

        # No timespans
        self.assertEqual(self.dlg._get_timespans(), ("0:00:00", "0:00:00"))

        # With timespans
        self.dlg = ClipDialog(cast("MainFrame", self.frame), ditem)

        option_timespan = ["--external-downloader-args", "-ss 70 -to 165"]
        self.dlg.download_item.options.extend(option_timespan)

        self.assertEqual(self.dlg._get_timespans(), ("0:01:10", "0:02:45"))
        self.assertEqual(len(self.dlg.download_item.options), 4)

        # self.dlg.clip_start.SetValue(wx.TimeSpan(0, 1, 10))  # "0:01:10"
        # self.dlg.clip_end.SetValue(wx.TimeSpan(0, 2, 45))  # "0:02:45"

    def test_get_timespans_no_exist(self):
        """With timespans no send (no exist)"""
        ditem = DownloadItem("url", ["-f", "flv"])
        self.dlg = ClipDialog(cast("MainFrame", self.frame), ditem)

        option_timespan = ["--external-downloader-args"]
        self.dlg.download_item.options.extend(option_timespan)

        self.assertEqual(self.dlg._get_timespans(), ("0:00:00", "0:00:00"))
        self.assertEqual(len(self.dlg.download_item.options), 2)

    def test_get_timespans_no_numbers(self):
        """With timespans like strings"""
        ditem = DownloadItem("url", ["-f", "flv"])
        self.dlg = ClipDialog(cast("MainFrame", self.frame), ditem)

        option_timespan = ["--external-downloader-args", "-ss ab -to xy"]
        self.dlg.download_item.options.extend(option_timespan)

        self.assertEqual(self.dlg._get_timespans(), ("0:00:00", "0:00:00"))
        self.assertEqual(len(self.dlg.download_item.options), 2)

    def test_clean_options(self):
        """External downloader options alway last options"""
        ditem = DownloadItem("url", ["-f", "flv"])
        options = [
            "--external-downloader",
            "ffmpeg",
            "--external-downloader-args",
            "-ss 70 -to 165",
        ]

        self.dlg = ClipDialog(cast("MainFrame", self.frame), ditem)
        self.dlg.download_item.options.extend(options)

        self.assertEqual(self.dlg.download_item.options, ["-f", "flv"] + options)

        self.dlg._clean_options()

        self.assertEqual(self.dlg.download_item.options, ["-f", "flv"])

    def test_clean_options_extra_args(self):
        """Clean options and extra args in the end"""
        ditem = DownloadItem("url", ["-f", "flv"])
        options = [
            "--external-downloader",
            "ffmpeg",
            "--external-downloader-args",
            "-ss 70 -to 165",
            "-R",
            "-keep-fragments",
            "-v",
        ]

        self.dlg = ClipDialog(cast("MainFrame", self.frame), ditem)
        self.dlg.download_item.options.extend(options)

        self.dlg._clean_options()

        self.assertEqual(
            self.dlg.download_item.options, ["-f", "flv", "-R", "-keep-fragments", "-v"]
        )


if __name__ == "__main__":
    unittest.main()



"""Contains test cases for the ListBoxComboPopup in widgets.py module."""

import sys
import unittest
from pathlib import Path

import wx

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))

from youtube_dl_gui.widgets import ListBoxComboPopup


class TestCustomComboBox(unittest.TestCase):
    """Test cases for the CustomComboBox widget."""

    def setUp(self) -> None:
        self.app = wx.App()
        self.frame = wx.Frame(None)

        self.combobox = wx.ComboCtrl(self.frame, size=(180, -1), style=wx.CB_READONLY)
        self._popup_ctrl = ListBoxComboPopup(self.combobox)
        self.combobox.SetPopupControl(self._popup_ctrl)

        self._lb_popup_ctr = self._popup_ctrl.GetControl()

        # Call directly the ListBoxWithHeaders methods
        self._lb_popup_ctr.add_header("Header")
        self._lb_popup_ctr.add_items([f"item{i}" for i in range(10)])

    def tearDown(self):
        self.frame.Destroy()
        del self.app

    def test_init(self):
        choices = ["item0", "item1", "item2"]
        combobox = wx.ComboCtrl(self.frame, size=(180, -1), style=wx.CB_READONLY)
        popup_ctrl = ListBoxComboPopup(combobox)
        combobox.SetPopupControl(popup_ctrl)

        lb_popup_ctr = popup_ctrl.GetControl()
        lb_popup_ctr.AppendItems(choices)

        popup_ctrl.SetStringSelection("item1")

        self.assertEqual(popup_ctrl.GetStringValue(), "item1")
        self.assertEqual(popup_ctrl.GetSelection(), 1)
        self.assertEqual(lb_popup_ctr.GetCount(), 3)

    # wx.ComboBox methods
    # Not all of them since most of them are calls to ListBoxWithHeaders
    # methods and we already have tests for those

    def test_is_list_empty_false(self):
        self.assertFalse(self._popup_ctrl.IsListEmpty())

    def test_is_list_empty_true(self):
        self._popup_ctrl.Clear()
        self.assertTrue(self._popup_ctrl.IsListEmpty())

    def test_is_text_empty_false(self):
        self.combobox.SetValue("somevalue")
        self.assertFalse(self.combobox.GetValue() == "")

    def test_is_text_empty_true(self):
        self.assertTrue(self.combobox.GetValue() == "")

    def test_set_selection_item(self):
        self._popup_ctrl.SetSelection(1)
        self.assertEqual(self._popup_ctrl.GetSelection(), 1)
        self.assertEqual(self.combobox.GetValue(), "item0")

    def test_set_selection_header(self):
        self._popup_ctrl.SetSelection(0)
        self.assertEqual(self._popup_ctrl.GetSelection(), wx.NOT_FOUND)
        self.assertEqual(self.combobox.GetValue(), "")

    def test_set_string_selection_item(self):
        self._popup_ctrl.SetStringSelection("item0")
        self.assertEqual(self._popup_ctrl.GetStringValue(), "item0")
        self.assertEqual(self.combobox.GetValue(), "item0")

    def test_set_string_selection_header(self):
        self._popup_ctrl.SetStringSelection("Header")
        self.assertEqual(self.combobox.GetStringSelection(), "")
        self.assertEqual(self.combobox.GetValue(), "")

    def test_set_string_selection_invalid_string(self):
        self._popup_ctrl.SetStringSelection("abcde")
        self.assertEqual(self.combobox.GetStringSelection(), "")
        self.assertEqual(self.combobox.GetValue(), "")

    # wx.ItemContainer methods

    def test_clear(self):
        self.combobox.SetValue("value")
        self._popup_ctrl.Clear()
        self.assertEqual(self._popup_ctrl.GetControl().GetCount(), 0)
        self.assertEqual(self.combobox.GetValue(), "")

    def test_append(self):
        self._popup_ctrl.GetControl().Append("item10")
        self.assertEqual(self._popup_ctrl.GetControl().GetCount(), 12)

    def test_append_items(self):
        self._popup_ctrl.GetControl().AppendItems(["item10", "item11"])
        self.assertEqual(self._popup_ctrl.GetControl().GetCount(), 13)

    def test_delete(self):
        self._popup_ctrl.GetControl().Delete(1)
        self.assertEqual(self._popup_ctrl.GetControl().GetString(1), "item1")

    # wx.TextEntry methods

    def test_get_value(self):
        self.combobox.SetValue("value")
        self.assertEqual(self.combobox.GetValue(), "value")


def main():
    unittest.main()


if __name__ == "__main__":
    main()



"""Contains test cases for the ButtonsChoiceDialog in widgets.py module."""

import sys
import unittest
from pathlib import Path

PATH = Path(__file__).parent
sys.path.insert(0, str(PATH.parent))


from tests.wtc import WidgetTestCase

from youtube_dl_gui.widgets import ButtonsChoiceDialog


class TestButtonsChoiceDialog(WidgetTestCase):
    """Test cases for the ButtonsChoiceDialog widget."""

    def test_init_dark(self):
        _dark_mode = True
        choices = ["Remove all", "Remove completed"]
        btn_choice_dlg = ButtonsChoiceDialog(
            self.frame, choices, "Test ButtonsChoiceDialog", "Test", _dark_mode
        )
        self.assertTrue(btn_choice_dlg._dark_mode)
        self.assertEqual(len(btn_choice_dlg.buttons), 3)


if __name__ == "__main__":
    unittest.main()



"""yt-dlg setup file.

Examples:
    Windows/Linux::

        python setup.py pyinstaller

    Linux::

        python setup.py install

    Build source distribution::

        python setup.py sdist

    Build platform distribution::

        python setup.py bdist

    Build the translations::

        python setup.py build_trans

    Build with updates disabled::

        python setup.py no_updates


    Basic steps of the setup::

        * Run pre-build tasks
        * Call setup handler based on OS & options
            * Set up hicolor icons (if supported by platform)
            * Set up fallback pixmaps icon (if supported by platform)
            * Set up package level pixmaps icons (*.png)
            * Set up package level i18n files (*.mo)
            * Set up scripts (executables) (if supported by platform)
        * Run setup

"""


import glob
import logging
import os
import re
import subprocess
import sys
from collections import defaultdict
from pathlib import Path

from setuptools import Command, setup


def load_vars(file=Path("youtube_dl_gui/info.py")):
    source = file.read_text(encoding="utf-8")
    _vars = defaultdict(lambda: "")

    # Docs file
    match = re.search(r'["]{3}(.*)["]{3}', source)
    if match:
        _vars["__doc__"] = f"{match.groups(0)[0]}"

    match = re.findall(r'([_]{2}[a-z]+[_]{2})\s+=\s+["\']{1}(.+)["\']{1}', source)
    if match:
        _vars.update(dict(match))
    return _vars


PYINSTALLER = len(sys.argv) >= 2 and sys.argv[1] == "pyinstaller"

vars_file = load_vars()
vars_file.update(load_vars(Path("youtube_dl_gui/version.py")))

try:
    from PyInstaller import compat as pyi_compat

    if pyi_compat.is_win:
        # noinspection PyUnresolvedReferences
        from PyInstaller.utils.win32.versioninfo import (
            FixedFileInfo,
            SetVersion,
            StringFileInfo,
            StringStruct,
            StringTable,
            VarFileInfo,
            VarStruct,
            VSVersionInfo,
        )
except ImportError:
    pyi_compat = None
    if PYINSTALLER:
        print("Cannot import pyinstaller", file=sys.stderr)
        exit(1)


DESCRIPTION = vars_file["__description__"]
LONG_DESCRIPTION = open("README.md", encoding="utf-8").read()


def on_windows():
    """Returns True if OS is Windows."""
    return os.name == "nt"


def version2tuple(commit=0):
    version_list = str(vars_file["__version__"]).split(".")

    if len(version_list) > 3:
        del version_list[3]

    _release = commit

    _year, _month, _day = tuple(map(int, version_list))
    return _year, _month, _day, _release


def version2str(commit=0):
    version_tuple = version2tuple(commit)
    return "%s.%s.%s.%s" % version_tuple


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# noinspection PyAttributeOutsideInit,PyArgumentList
class BuildTranslations(Command):
    description = "Build the translation files"
    user_options = []

    def initialize_options(self):
        self.search_pattern = None

    def finalize_options(self):
        self.search_pattern = os.path.join(
            vars_file["__packagename__"],
            "locale",
            "*",
            "LC_MESSAGES",
            "youtube_dl_gui.po",
        )

    def run(self):
        import polib

        po_file = ""

        try:
            for po_file in glob.glob(self.search_pattern):
                mo_file = po_file.replace(".po", ".mo")
                po = polib.pofile(po_file)

                logger.info(f"Building MO file for '{po_file}'")
                po.save_as_mofile(mo_file)
        except OSError as error:
            logger.error(f"{error}, exiting...")
            sys.exit(1)


# noinspection PyAttributeOutsideInit,PyArgumentList
class BuildNoUpdate(Command):
    description = "Build with updates disabled"
    user_options = []

    def initialize_options(self):
        self.build_lib = os.path.dirname(os.path.abspath(__file__))

    def finalize_options(self):
        pass

    def run(self):
        self.__disable_updates()

    def __disable_updates(self):
        lib_dir = os.path.join(self.build_lib, vars_file["__packagename__"])
        target_file = "optionsmanager.py"
        # Options file should be available from previous build commands
        optionsfile = os.path.join(lib_dir, target_file)

        with open(optionsfile) as input_file:
            data = input_file.readlines()

        if data is None:
            logger.error("Building with updates disabled failed!")
            sys.exit(1)

        for index, line in enumerate(data):
            if '"disable_update": False' in line:
                logger.info("Disabling updates...")
                data[index] = line.replace("False", "True")
                break

        with open(optionsfile, "w") as output_file:
            output_file.writelines(data)


class BuildPyinstallerBin(Command):
    description = "Build the executable"
    user_options = []
    version_file = None
    if pyi_compat and pyi_compat.is_win:
        version_file = VSVersionInfo(
            ffi=FixedFileInfo(
                filevers=version2tuple(),
                prodvers=version2tuple(),
                mask=0x3F,
                flags=0x0,
                OS=0x4,
                fileType=0x1,
                subtype=0x0,
                date=(0, 0),
            ),
            kids=[
                VarFileInfo([VarStruct("Translation", [0, 1200])]),
                StringFileInfo(
                    [
                        StringTable(
                            "000004b0",
                            [
                                StringStruct("CompanyName", vars_file["__mcontact__"]),
                                StringStruct("FileDescription", DESCRIPTION),
                                StringStruct("FileVersion", version2str()),
                                StringStruct("InternalName", "yt-dlg.exe"),
                                StringStruct(
                                    "LegalCopyright",
                                    f"{vars_file['__projecturl__']}LICENSE",
                                ),
                                StringStruct("OriginalFilename", "yt-dlg.exe"),
                                StringStruct("ProductName", vars_file["__appname__"]),
                                StringStruct("ProductVersion", version2str()),
                            ],
                        )
                    ]
                ),
            ],
        )

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self, version=version_file):
        """Run pyinstaller"""
        path_sep = ";" if on_windows() else ":"
        subprocess.run(
            [
                "pyinstaller",
                "-w",
                "-F",
                f"--icon={vars_file['__packagename__']}/data/pixmaps/youtube-dl-gui.ico",
                "--add-data="
                + vars_file["__packagename__"]
                + "/data"
                + path_sep
                + vars_file["__packagename__"]
                + "/data",
                "--add-data="
                + vars_file["__packagename__"]
                + "/locale"
                + path_sep
                + vars_file["__packagename__"]
                + "/locale",
                "--exclude-module=tests",
                "--name=yt-dlg",
                f"{vars_file['__packagename__']}/__main__.py",
            ]
        )

        if version:
            SetVersion("./dist/yt-dlg.exe", version)


cmdclass = {
    "build_trans": BuildTranslations,
    "no_updates": BuildNoUpdate,
}

# Add pixmaps icons (*.png) & i18n files
package_data = {
    vars_file["__packagename__"]: ["data/pixmaps/*.png", "locale/*/LC_MESSAGES/*.mo"]
}


def setup_linux():
    """Setup params for Linux"""
    data_files_linux = []
    # Add hicolor icons
    for path in glob.glob("youtube_dl_gui/data/icons/hicolor/*x*"):
        size = os.path.basename(path)

        dst = f"share/icons/hicolor/{size}/apps"
        src = f"{path}/apps/youtube-dl-gui.png"

        data_files_linux.append((dst, [src]))
    data_files_linux.extend(
        (
            (
                "share/pixmaps",
                ["youtube_dl_gui/data/pixmaps/youtube-dl-gui.png"],
            ),
            ("share/man/man1", ["youtube-dl-gui.1"]),
        )
    )

    return {
        "data_files": data_files_linux,
        "package_data": package_data,
    }


def setup_windows():
    """Setup params for Windows"""
    return {
        "package_data": package_data,
    }


params = {}

if PYINSTALLER:
    cmdclass.update({"pyinstaller": BuildPyinstallerBin})
else:
    params = setup_windows() if on_windows() else setup_linux()
    params["entry_points"] = {
        "console_scripts": [
            f"{vars_file['__appname__']} = {vars_file['__packagename__']}.app:main"
        ]
    }


if __name__ == "__main__":
    setup(
        name=f"{vars_file['__appname__']}".replace("-", "_"),
        version=vars_file["__version__"],
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        url=vars_file["__projecturl__"],
        download_url=f"{vars_file['__githuburl__']}releases/latest",
        project_urls={
            "Source": vars_file["__githuburl__"],
            "Tracker": f"{vars_file['__githuburl__']}issues",
            "Funding": f"{vars_file['__projecturl__']}donate.html",
        },
        author=vars_file["__author__"],
        author_email=vars_file["__contact__"],
        maintainer=vars_file["__maintainer__"],
        maintainer_email=vars_file["__mcontact__"],
        license=vars_file["__license__"],
        packages=[vars_file["__packagename__"]],
        install_requires=[
            "pypubsub>=4.0.3",
            "wxPython<=4.2.1a1,>=4.0.7.post2",
        ],
        extras_require={
            "binaries": ["polib>=1.1.0", "pyinstaller<=5.8.0,>=3.6"],
        },
        python_requires=">=3.7",
        classifiers=[
            "Topic :: Multimedia :: Video",
            "Development Status :: 5 - Production/Stable",
            "Environment :: MacOS X :: Cocoa",
            "Environment :: Win32 (MS Windows)",
            "Environment :: X11 Applications :: GTK",
            "License :: Public Domain",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: Implementation :: CPython",
        ],
        cmdclass=cmdclass,
        **params,
    )



"""yt-dlg module for managing the download process.

This module is responsible for managing the download process
and update the GUI interface.

Attributes:
    MANAGER_PUB_TOPIC (str): Publisher subscription topic of the
        DownloadManager thread.

    WORKER_PUB_TOPIC (str): Publisher subscription topic of the
        Worker thread.

Note:
    It's not the actual module that downloads the urls
    thats the job of the 'downloaders' module.

"""
from __future__ import annotations

import time
from pathlib import Path
from threading import RLock, Thread
from typing import TYPE_CHECKING, Any, Callable

import wx

# noinspection PyPep8Naming
from pubsub import pub as Publisher

from .downloaders import YoutubeDLDownloader
from .parsers import OptionsParser
from .updatemanager import UpdateThread
from .utils import YOUTUBEDL_BIN, format_bytes, to_bytes

if TYPE_CHECKING:
    from .logmanager import LogManager
    from .mainframe import MainFrame
    from .optionsmanager import OptionsManager

MANAGER_PUB_TOPIC = "dlmanager"
WORKER_PUB_TOPIC = "dlworker"

_SYNC_LOCK = RLock()


def synchronized(lock: RLock) -> Callable[..., Any]:
    """Decorator that adds thread synchronization to a function"""

    def _decorator(func: Callable) -> Callable[..., Any]:
        def _wrapper(*args, **kwargs) -> Any:
            lock.acquire()
            ret_value: Any = func(*args, **kwargs)
            lock.release()
            return ret_value

        return _wrapper

    return _decorator


class DownloadItem:
    """Object that represents a download.

    Attributes:
        STAGES (tuple): Main stages of the download item.

        ACTIVE_STAGES (tuple): Sub stages of the 'Active' stage.

        COMPLETED_STAGES (tuple): Sub stages of the 'Completed' stage.

        ERROR_STAGES (tuple): Sub stages of the 'Error' stage.

    Args:
        url (string): URL that corresponds to the download item.

        options (list): Options list to use during the download phase.

    """

    STAGES = ("Queued", "Active", "Paused", "Completed", "Error")

    ACTIVE_STAGES = ("Pre Processing", "Downloading", "Post Processing")

    COMPLETED_STAGES = ("Finished", "Warning", "Already Downloaded")

    ERROR_STAGES = ("Error", "Stopped", "Filesize Abort")

    def __init__(self, url: str, options: list[str]):
        self.url: str = url
        self.options: list[str] = options
        self.object_id: int = hash(url + str(options))
        self._stage: str = self.STAGES[0]
        self.path: str = ""
        self.filenames: list[str] = []
        self.extensions: list[str] = []
        self.filesizes: list[float] = []
        self.progress_stats: dict[str, str] = {}
        self.playlist_index_changed: bool = False

        self.reset()

    @property
    def stage(self) -> str:
        return self._stage

    @stage.setter
    def stage(self, value: str) -> None:
        if value not in self.STAGES:
            raise ValueError(value)

        if value == "Active":
            self.progress_stats["status"] = self.ACTIVE_STAGES[0]
        elif value == "Completed":
            self.progress_stats["status"] = self.COMPLETED_STAGES[0]
        elif value == "Error":
            self.progress_stats["status"] = self.ERROR_STAGES[0]

        elif value in {"Queued", "Paused"}:
            self.progress_stats["status"] = value
        self._stage = value

    def _init_filename_sizes_extensions(self) -> None:
        self.filenames = []
        self.extensions = []
        self.filesizes = []

    def reset(self) -> None:
        if self._stage == self.STAGES[1]:
            raise RuntimeError("Cannot reset an 'Active' item")

        self._stage = self.STAGES[0]
        self.path = ""
        self._init_filename_sizes_extensions()

        self.default_values: dict[str, str] = {
            "filename": self.url,
            "extension": "-",
            "filesize": "-",
            "percent": "0%",
            "speed": "-",
            "eta": "-",
            "status": self.stage,
            "playlist_size": "",
            "playlist_index": "",
        }

        self.progress_stats = dict(self.default_values)
        # Keep track when the 'playlist_index' changes
        self.playlist_index_changed = False

    def get_files(self) -> list[str]:
        """Returns a list that contains all the system files bind to this object."""
        return [
            str(Path(self.path) / Path(item + self.extensions[index]))
            for index, item in enumerate(self.filenames)
        ]

    def update_stats(self, stats_dict: dict[str, Any]) -> None:
        """Updates the progress_stats dict from the given dictionary."""
        assert isinstance(stats_dict, dict)

        for key in stats_dict:
            if key in self.progress_stats:
                value = stats_dict.get(key)

                self.progress_stats[key] = value or self.default_values[key]
        # Extract extra stuff
        if "playlist_index" in stats_dict:
            self.playlist_index_changed = True

        if "filename" in stats_dict:
            # Reset filenames, extensions & filesizes lists when changing playlist item
            if self.playlist_index_changed:
                self._init_filename_sizes_extensions()
                self.playlist_index_changed = False

            self.filenames.append(stats_dict["filename"])

        if "extension" in stats_dict:
            self.extensions.append(stats_dict["extension"])

        if "path" in stats_dict:
            self.path = stats_dict["path"]

        if (
            "filesize" in stats_dict
            and stats_dict["percent"] == "100%"
            and len(self.filesizes) < len(self.filenames)
        ):
            filesize = stats_dict["filesize"].lstrip("~")  # HLS downloader etc
            self.filesizes.append(to_bytes(filesize))

        if "status" in stats_dict:
            # If we are post processing try to calculate the size of
            # the output file since youtube-dl does not
            if (
                stats_dict["status"] == self.ACTIVE_STAGES[2]
                and len(self.filesizes) == 2
            ):
                post_proc_filesize = self.filesizes[0] + self.filesizes[1]

                self.filesizes.append(post_proc_filesize)
                self.progress_stats["filesize"] = format_bytes(post_proc_filesize)

            self._set_stage(stats_dict["status"])

    def _set_stage(self, status: str) -> None:
        if status in self.ACTIVE_STAGES:
            self._stage = self.STAGES[1]

        if status in self.COMPLETED_STAGES:
            self._stage = self.STAGES[3]

        if status in self.ERROR_STAGES:
            self._stage = self.STAGES[4]

    def __eq__(self, other: object) -> bool:
        return (
            self.object_id == other.object_id  # type: ignore[return-value]
            if isinstance(other, DownloadItem)
            else NotImplemented
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.url},{self.options})>"


class DownloadList:
    """List like data structure that contains DownloadItems.

    Args:
        items (list): List that contains DownloadItems.

    """

    def __init__(self, items: list[DownloadItem] | None = None):
        assert isinstance(items, list) or items is None

        self._items_dict: dict[int, DownloadItem] = {}  # Speed up lookup
        self._items_list: list[int] = []  # Keep the sequence

        if items:
            self._items_list = [item.object_id for item in items]
            self._items_dict = {item.object_id: item for item in items}

    @synchronized(_SYNC_LOCK)
    def clear(self) -> None:
        """Removes all the items from the list even the 'Active' ones."""
        self._items_list = []
        self._items_dict = {}

    @synchronized(_SYNC_LOCK)
    def insert(self, item: DownloadItem) -> None:
        """Inserts the given item to the list. Does not check for duplicates."""
        self._items_list.append(item.object_id)
        self._items_dict[item.object_id] = item

    @synchronized(_SYNC_LOCK)
    def remove(self, object_id: int | None) -> bool:
        """Removes an item from the list.

        Removes the item with the corresponding object_id from
        the list if the item is not in 'Active' state.

        Returns:
            True on success else False.

        """
        assert object_id is not None

        item = self._items_dict[object_id]

        if item and item.stage != "Active":
            self._items_list.remove(object_id)
            del self._items_dict[object_id]

            return True
        return False

    @synchronized(_SYNC_LOCK)
    def fetch_next(self) -> DownloadItem | None:
        """Returns the next queued item on the list.

        Returns:
            Next queued item or None if no other item exist.

        """
        for object_id in self._items_list:
            cur_item = self._items_dict[object_id]
            if cur_item.stage == "Queued":
                return cur_item

        return None

    @synchronized(_SYNC_LOCK)
    def move_up(self, object_id: int):
        """Moves the item with the corresponding object_id up to the list."""
        index: int = self._items_list.index(object_id)

        if index > 0:
            self._swap(index, index - 1)
            return True

        return False

    @synchronized(_SYNC_LOCK)
    def move_down(self, object_id: int):
        """Moves the item with the corresponding object_id down to the list."""
        index: int = self._items_list.index(object_id)

        if index < (len(self._items_list) - 1):
            self._swap(index, index + 1)
            return True

        return False

    @synchronized(_SYNC_LOCK)
    def get_item(self, object_id: int | None) -> DownloadItem | None:
        """Returns the DownloadItem with the given object_id."""
        assert object_id is not None
        return self._items_dict.get(object_id, None)

    @synchronized(_SYNC_LOCK)
    def has_item(self, object_id: int) -> bool:
        """Returns True if the given object_id is in the list else False."""
        return object_id in self._items_list

    @synchronized(_SYNC_LOCK)
    def get_items(self) -> list[DownloadItem]:
        """Returns a list with all the items."""
        return [self._items_dict[object_id] for object_id in self._items_list]

    @synchronized(_SYNC_LOCK)
    def change_stage(self, object_id: int, new_stage: str) -> None:
        """Change the stage of the item with the given object_id."""
        self._items_dict[object_id].stage = new_stage

    @synchronized(_SYNC_LOCK)
    def index(self, object_id: int) -> int:
        """Get the zero based index of the item with the given object_id."""
        if object_id in self._items_list:
            return self._items_list.index(object_id)
        return -1

    @synchronized(_SYNC_LOCK)
    def __len__(self) -> int:
        return len(self._items_list)

    @synchronized(_SYNC_LOCK)
    def __repr__(self) -> str:
        return str(dict(self._items_dict.items()))

    def _swap(self, index1: int, index2: int) -> None:
        self._items_list[index1], self._items_list[index2] = (
            self._items_list[index2],
            self._items_list[index1],
        )


class DownloadManager(Thread):
    """Manages the download process.

    Attributes:
        WAIT_TIME (float): Time in seconds to sleep.

    Args:
        parent (mainframe.MainFrame): Main Frame

        download_list (DownloadList): List that contains items to download.

        opt_manager (optionsmanager.OptionsManager): Object responsible for
            managing the yt-dlg options.

        log_manager (logmanager.LogManager): Object responsible for writing
            errors to the log.

        daemon (bool): If Thread is daemonic

    """

    WAIT_TIME = 0.1

    def __init__(
        self,
        parent: MainFrame,
        download_list: DownloadList,
        opt_manager: OptionsManager,
        log_manager: LogManager | None = None,
        daemon=False,
    ):
        super().__init__()
        self.parent = parent
        self.opt_manager = opt_manager
        self.log_manager = log_manager
        self.download_list = download_list

        self._time_it_took: float = 0
        self._successful = 0
        self._running = True

        # Init the custom workers thread pool
        self._workers = [
            Worker(opt_manager, self._youtubedl_path(), log_manager, worker=worker)
            for worker in range(1, int(opt_manager.options["workers_number"]) + 1)
        ]

        self.name = "DownloadManager"
        self.daemon = daemon
        self.start()

    @property
    def successful(self) -> int:
        """Returns number of successful downloads."""
        return self._successful

    @property
    def time_it_took(self) -> float:
        """Returns time(seconds) it took for the download process
        to complete."""
        return self._time_it_took

    def run(self) -> None:
        if not self.opt_manager.options["disable_update"]:
            self._check_youtubedl()

        self._time_it_took = time.time()

        # TODO: Use threading.Condition
        while self._running:
            item: DownloadItem | None = self.download_list.fetch_next()

            if item is not None:
                worker = self._get_worker()

                if worker is not None:
                    worker.download(item.url, item.options, item.object_id)
                    self.download_list.change_stage(item.object_id, "Active")

            if item is None and self._jobs_done():
                break

            time.sleep(self.WAIT_TIME)

        # Close all the workers and collect
        for worker in self._workers:
            worker.close()
            self._successful += worker.successful

        self._time_it_took = time.time() - self._time_it_took

        if not self._running:
            self._talk_to_gui("closed")
        else:
            self._talk_to_gui("finished")

    def active(self) -> int:
        """Returns number of active items."""
        return len(self.download_list)

    def stop_downloads(self) -> None:
        """Stop the download process. Also send 'closing'
        signal back to the GUI.

        Note:
            It does NOT kill the workers thats the job of the
            clean up task in the run() method.

        """
        self._talk_to_gui("closing")
        self._running = False

    def send_to_worker(self, data: dict[str, Any]) -> None:
        """Send data to the Workers.

        Args:
            data (dict): Python dictionary that holds the 'index'
            which is used to identify the Worker thread and the data which
            can be any of the Worker's class valid data. For a list of valid
            data keys see __init__() under the Worker class.

        """
        if "index" in data:
            for worker in self._workers:
                if worker.has_index(data["index"]):
                    worker.update_data(data)

    @staticmethod
    def _talk_to_gui(signal: str, data: dict[str, Any] | None = None) -> None:
        """Send data back to the GUI using wxCallAfter and wxPublisher.

        Args:
            signal (string): Unique signal string that informs the GUI for the
                download process.

        Note:
            DownloadManager supports 4 signals.
                1) closing: The download process is closing.
                2) closed: The download process has closed.
                3) finished: The download process was completed normally.
                4) report_active: Signal the gui to read the number of active
                    downloads using the active() method.

        """
        if wx.GetApp() is not None:
            wx.CallAfter(
                Publisher.sendMessage, MANAGER_PUB_TOPIC, signal=signal, data=data
            )

    def _check_youtubedl(self) -> None:
        """Check if youtube-dl binary exists. If not try to download it."""
        ytdl_path = self._youtubedl_path()
        if not Path(ytdl_path).exists() and self.parent.update_thread is None:
            self.parent.update_thread = UpdateThread(self.opt_manager, True)
            self.parent.update_thread.join()
            self.parent.update_thread = None

    def _get_worker(self) -> Worker | None:
        return next((worker for worker in self._workers if worker.available()), None)

    def _jobs_done(self) -> bool:
        """Returns True if the workers have finished their jobs else False."""
        return all(worker.available() for worker in self._workers)

    def _youtubedl_path(self) -> str:
        """Returns the path to youtube-dl binary."""
        cli_backend: str = self.opt_manager.options.get("cli_backend", YOUTUBEDL_BIN)
        return str(Path(self.opt_manager.options["youtubedl_path"]) / Path(cli_backend))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.download_list})>"


class Worker(Thread):
    """Simple worker which downloads the given url using a downloader
    from the downloaders.py module.

    Attributes:
        WAIT_TIME (float): Time in seconds to sleep.

        worker_count (int): Numer of worker threads

    Args:
        opt_manager (optionsmanager.OptionsManager): Check DownloadManager
            description.

        youtubedl_path (str): Absolute path to youtube-dl binary.

        log_manager (logmanager.LogManager): Check DownloadManager
            description.

        worker (int): Worker thread number

    Note:
        For available data keys see self._data under the __init__() method.

    """

    WAIT_TIME = 0.1
    worker_count = 0

    def __init__(
        self,
        opt_manager: OptionsManager,
        youtubedl_path: str,
        log_manager: LogManager | None = None,
        worker: int | None = None,
    ):
        super().__init__()
        # Use Daemon ?
        # self.setDaemon(True)
        Worker.worker_count += 1
        self.opt_manager = opt_manager
        self.log_manager = log_manager
        self.worker = worker or Worker.worker_count
        self.setName(f"Worker_{worker}")

        self._downloader = YoutubeDLDownloader(
            youtubedl_path, self._data_hook, self._log_data
        )
        self._options_parser = OptionsParser()
        self._successful = 0
        self._running = True
        self._options: list[str] | None = None

        self._wait_for_reply = False

        self._data: dict[str, Any] = {
            "playlist_index": None,
            "playlist_size": None,
            "new_filename": None,
            "extension": None,
            "filesize": None,
            "filename": None,
            "percent": None,
            "status": None,
            "index": None,
            "speed": None,
            "path": None,
            "eta": None,
            "url": None,
        }

        self.start()

    def run(self) -> None:
        # TODO: Use threading.Condition
        while self._running:
            if self._data.get("url"):
                ret_code = self._downloader.download(self._data["url"], self._options)

                if ret_code in [
                    YoutubeDLDownloader.OK,
                    YoutubeDLDownloader.ALREADY,
                    YoutubeDLDownloader.WARNING,
                ]:
                    self._successful += 1

                self._reset()

            time.sleep(self.WAIT_TIME)

        # Call the destructor function of YoutubeDLDownloader object
        self._downloader.close()

    # noinspection PyIncorrectDocstring
    def download(self, url: str, options: list[str], object_id: int) -> None:
        # noinspection PyUnresolvedReferences
        """Download given item.

        Args:
            item (dict): Python dictionary that contains two keys.
                The url and the index of the corresponding row in which
                the worker should send back the information about the
                download process.

        """
        self._data["url"] = url
        self._options = options
        self._data["index"] = object_id

    def stop_download(self) -> None:
        """Stop the download process of the worker."""
        self._downloader.stop()

    def close(self) -> None:
        """Kill the worker after stopping the download process."""
        self.stop_download()
        self._running = False

    def available(self) -> bool:
        """Return True if the worker has no job else False."""
        return self._data["url"] is None

    def has_index(self, index) -> bool:
        """Return True if index is equal to self._data['index'] else False."""
        return self._data["index"] == index

    def update_data(self, data: dict[str, Any]) -> None:
        """Update self._data from the given data."""
        if self._wait_for_reply:
            # Update data only if a receive request has been issued
            self._data.update(data)

            self._wait_for_reply = False

    @property
    def successful(self) -> int:
        """Return the number of successful downloads for current worker."""
        return self._successful

    def _reset(self) -> None:
        """Reset self._data back to the original state."""
        for key in self._data:
            self._data[key] = None

    def _log_data(self, data: str) -> None:
        """Callback method for self._downloader.

        This method is used to write the given data in a synchronized way
        to the log file using the self.log_manager.

        Args:
            data (str): String to write to the log file.

        """
        if self.log_manager is not None:
            self.log_manager.log(data)

    def _data_hook(self, data: dict[str, Any]) -> None:
        """Callback method for self._downloader.

        This method updates self._data and sends the updates back to the
        GUI using the self._talk_to_gui() method.

        Args:
            data (dict): Python dictionary which contains information
                about the download process. For more info see the
                extract_data() function under the `downloaders` module.

        """
        self._talk_to_gui("send", data)

    def _talk_to_gui(self, signal: str, data: dict[str, Any]) -> None:
        """Communicate with the GUI using wxCallAfter and Publisher.

        Send/Ask data to/from the GUI. Note that if the signal is 'receive'
        then the Worker will wait until it receives a reply from the GUI.

        Args:
            signal (str): Unique string that informs the GUI about the
                communication procedure.

            data (dict): Python dictionary which holds the data to be sent
                back to the GUI. If the signal is 'send' then the dictionary
                contains the updates for the GUI (e.g. percentage, eta). If
                the signal is 'receive' then the dictionary contains exactly
                three keys. The 'index' (row) from which we want to retrieve
                the data, the 'source' which identifies a column in the
                wxListCtrl widget and the 'dest' which tells the wxListCtrl
                under which key to store the retrieved data.

        Note:
            Worker class supports 2 signals.
                1) send: The Worker sends data back to the GUI
                         (e.g. Send status updates).
                2) receive: The Worker asks data from the GUI
                            (e.g. Receive the name of a file).

        Structure:
            ('send', {'index': <item_row>, data_to_send*})

            ('receive', {'index': <item_row>, 'source': 'source_key', 'dest': 'destination_key'})

        """
        data["index"] = self._data["index"]

        if signal == "receive":
            self._wait_for_reply = True

        if wx.GetApp() is not None:
            wx.CallAfter(
                Publisher.sendMessage, WORKER_PUB_TOPIC, signal=signal, data=data
            )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(Worker_{self.worker})>"
